"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
Last modified: 3 March 2006

FabTurtle: Draws vector graphics using a LOGO like interface.
           Adjust machine parameters (force, speed, rate) at any time.
           Export to SVG or straight to Epilog or Camm formats.
"""

from math import *

"""
shapes: [{'x': [...], 'y': [...]}, ...]
"""

class FabTurtle:
   def __init__(self, commands = []):
      self.state = {'pos': [0, 0], 'dir': 0, 'shapes': [], 'pen': False}
      self.command(commands)

   def command(self, commands):
      """
      Moves the turtle around

      The commands argument is a heterogeneous list of strings and parameters as below:
        PU           - pen up
        PD           - pen down
        ROT #        - rotate clockwise (radians)
        MOV #        - move forward
        ANG #        - set angle (radians)
        POS # #      - set position (x, y)
        PARAM object - set parameters
                       ['force', #]
                       ['speed', #]
                       ['rate', #]
      """

      ind = 0
      while ind < len(commands):
         command = commands[ind]
         ind += 1
         if command == 'PU':
            self.state['pen'] = False
            if (len(self.state['shapes']) > 0) and (len(self.state['shapes'][-1]) <= 1):
               self.state['shapes'].pop()
         elif command == 'PD':
            if not self.state['pen']:
               self.state['pen'] = True
            self.state['shapes'].append({'type': 'curve', 'x': [self.state['pos'][0]], 'y': [self.state['pos'][1]]})
         elif command == 'ROT':
            self.state['dir'] += commands[ind]
            ind += 1
         elif command == 'MOV':
            self.state['pos'][0] += commands[ind] * cos(self.state['dir'])
            self.state['pos'][1] += commands[ind] * sin(self.state['dir'])
            ind += 1
            if self.state['pen']:
               self.state['shapes'][-1]['x'].append(self.state['pos'][0])
               self.state['shapes'][-1]['y'].append(self.state['pos'][1])
         elif command == 'ANG':
            self.state['dir'] = commands[ind]
            ind += 1
         elif command == 'POS':
            self.state['pos'] = [commands[ind], commands[ind+1]]
            ind += 2
            if self.state['pen']:
               self.state['shapes'][-1]['x'].append(self.state['pos'][0])
               self.state['shapes'][-1]['y'].append(self.state['pos'][1])
         elif command == 'PARAM':
            if (len(self.state['shapes']) > 0) and (len(self.state['shapes'][-1]) <= 1):
               self.state['shapes'].pop()
            self.state['shapes'].append({'type': 'parameter', 'data': commands[ind]})
            ind += 1
            self.state['shapes'].append({'type': 'curve', 'x': [self.state['pos'][0]], 'y': [self.state['pos'][1]]})
         else:
            print 'Error: Unknown command "' + command + '"'
      
   def write_svg(self, filename, units = 'mm', scale = 1., xoff = 0., yoff = 0.):
      """
      Writes a shapes structure to a Scalar Vector Graphics (SVG) file
      Warning: parameter settings in the shapes structure are ignored

      Arguments:
        filename: path to output file [string]
        units: 'mm' | 'in' [string]
        scale, xoff, yoff: scale and shift the shape [float]
      """

      # Find bounding box
      HUGE = 1e10
      xmin = HUGE
      ymin = HUGE
      xmax = -HUGE
      ymax = -HUGE
      for s in range(len(self.state['shapes'])):
         if self.state['shapes'][s]['type'] == 'curve':
            if len(self.state['shapes'][s]['x']) > 1:
               for i in range(len(self.state['shapes'][s]['x'])):
                  xmin = min(xmin, self.state['shapes'][s]['x'][i]*scale+xoff)
                  xmax = max(xmax, self.state['shapes'][s]['x'][i]*scale+xoff)
                  ymin = min(ymin, self.state['shapes'][s]['y'][i]*scale+yoff)
                  ymax = max(ymax, self.state['shapes'][s]['y'][i]*scale+yoff)

      # Open output file and write header
      outfile = open(filename, "w")
      outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n')
      outfile.write('<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.0//EN" "http://www.w3#.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">\n')
      outfile.write('<svg width="%f%s" height="%f%s" viewBox="%f %f %f %f">\n'\
                    %(xmax-xmin,units,ymax-ymin,units,xmin,ymin,xmax-xmin,ymax-ymin))
      outfile.write('<g style="stroke:rgb(0,0,0);fill:none">\n')

      # Write the shapes (parameter commands are ignored)
      for s in range(len(self.state['shapes'])):
         if self.state['shapes'][s]['type'] == 'curve':
            x = self.state['shapes'][s]['x']
            y = self.state['shapes'][s]['y']
            if len(x) > 1:
               outfile.write(' <polyline points="%f,%f'%(x[0]*scale+xoff, y[0]*scale+yoff))
               for point in range(1, len(x)):
                  outfile.write(' %f,%f'%(x[point]*scale+xoff, y[point]*scale+yoff))
               outfile.write('" style="fill:none"/>\n')

      # Write footer and close file
      outfile.write('</g>\n')
      outfile.write('</svg>\n')
      outfile.close()

   def write_epi(self, filename, rate, power, speed, scale = 1., xoff = 0., yoff = 0., bedheight = 12., autofocus = 1):
      """
      Writes a shapes structure to a file for printing to the Epilog
      Possible shape parameter settings: rate, force, speed

      Arguments:
        filename: path to output file [string]
        rate, power, speed: initial Epilog settings [int]
        scale, xoff, yoff: scale and shift the shape [float]
        bedheight: the maximum y-coordinate of the Epilog [float]
        autofocus: 0 (off) / 1 (on) [int]
      """

      dpi = 600. # TODO: figure out where to write this in the header and make it an argument
   
      def transform(x, y):
         x = int(dpi*(x*scale+xoff))
         y = int(dpi*(bedheight-(y*scale+yoff)))
         return (x, y)

      # Open output file and write header
      outfile = open(filename, "w")
      outfile.write("%-12345X@PJL JOB NAME=" +
                    filename +
                    "\r\nE@PJL ENTER LANGUAGE=PCL\r\n&y" +
                    str(autofocus) +
                    "A&l0U&l0Z&u600D*p0X*p0Y*t600R*r0F&y50P&z50S*r6600T*r5100S*r1A*rC%1BIN;XR" +
                    str(rate) + ";YP" + str(power) + ";ZS" + str(speed) + ";")

      for s in range(len(self.state['shapes'])):
         if self.state['shapes'][s]['type'] == 'curve': # Output shape
            xvec = self.state['shapes'][s]['x']
            yvec = self.state['shapes'][s]['y']
            if len(xvec) > 1:
               (x, y) = transform(xvec[0], yvec[0])
               outfile.write('PU%i,%i;'%(x,y))
               for point in range(1, len(xvec)):
                  (x, y) = transform(xvec[point], yvec[point])
                  outfile.write('PD%i,%i;'%(x,y))
         elif self.state['shapes'][s]['type'] == 'parameter': # Adjust parameter
            if self.state['shapes'][s]['data'][0] == 'rate':
               outfile.write('XR%i;'%self.state['shapes'][s]['data'][1])
            elif self.state['shapes'][s]['data'][0] == 'force':
               outfile.write('YP%i;'%self.state['shapes'][s]['data'][1])
            elif self.state['shapes'][s]['data'][0] == 'speed':
               outfile.write('ZS%i;'%self.state['shapes'][s]['data'][1])

      # write footer and close file
      outfile.write("%0B%1BPUE%-12345X@PJL EOJ \r\n")
      outfile.close()

   def write_camm(self, filename, force, speed, scale = 1., xoff = 0., yoff = 0.):
      """
      Writes a shapes structure to a file for printing to the Camm
      Possible shape parameter settings: force, speed

      Arguments:
        filename: path to output file [string]
        force, speed: initial Camm settings [int]
        scale, xoff, yoff: scale and shift the shape [float]
      """

      dpi = 1000. # TODO: figure out where to write this in the header and make it an argument
   
      def transform(x, y):
         x = int(dpi*(x*scale+xoff))
         y = int(dpi*(y*scale+yoff))
         return (x, y)

      # Open output file and write header
      outfile = open(filename, "w")
      outfile.write("PA;PA;!ST1;VS" + str(speed) + ";PU;!FS" + str(force) + ";")
   
      for s in range(len(self.state['shapes'])):
         if self.state['shapes'][s]['type'] == 'curve': # Write shape
            xvec = self.state['shapes'][s]['x']
            yvec = self.state['shapes'][s]['y']
            if len(xvec) > 1:
               (x, y) = transform(xvec[0], yvec[0])
               outfile.write('PU%i,%i;'%(x,y))
               for point in range(1, len(xvec)):
                  (x, y) = transform(xvec[point], yvec[point])
                  outfile.write('PD%i,%i;'%(x,y))
         elif self.state['shapes'][s]['type'] == 'parameter': # Adjust parameter
            if self.state['shapes'][s]['data'][0] == 'force':
               outfile.write('PU;!FS%i;'%self.state['shapes'][s]['data'][1])
            elif self.state['shapes'][s]['data'][0] == 'speed':
               outfile.write('VS%i;'%self.state['shapes'][s]['data'][1])

      # Write footer and close file
      outfile.write("PU0,0;")
      outfile.close()
