"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division
import re

def epi2paths(commands, dpi = 600.):
    """
    Build a list of paths from Epilog Laser Cutter commands.
    Each path is {'params': {'rate':, 'speed':, 'power':},
                  'path': [[x,y], [x,y], ...]}
    """

    # Create command list
    start = commands.find('%1BIN;')
    if start == -1:
        start = None
    else:
        start += 6
    stop  = commands.rfind('%0B')
    if stop == -1:
        stop = None
    else:
        stop -= 1
    commands = commands[start:stop]
    commands = re.split('\s*;\s*', commands)

    # Build paths
    pen = {'x': 0, 'y': 0,
           'rate': 2500, 'power': 100, 'speed': 100,
           'down': False}
    paths = []
    for command in commands:
        if command == '': continue
        if command[0] == 'P':
            if (command[1] == 'D') and not pen['down']:
                # First PD: save initial parameters
                paths.append({'params': {'rate':  pen['rate'],
                                         'power': pen['power'],
                                         'speed': pen['speed']},
                              'path': [[pen['x']/dpi, pen['y']/dpi]]})
                pen['down'] = True
            if command[2:] != '': # Disregard PU; commands with no coordinates
                temp = command[2:].split(',')
                pen['x'] = int(temp[0])
                pen['y'] = int(temp[1])
            if command[1] == 'D': # PD
                paths[-1]['path'].append([pen['x']/dpi, pen['y']/dpi])
            else: # PU
                if (len(paths) > 0) and (len(paths[-1]['path']) < 2):
                    del paths[-1]
                pen['down'] = False
        elif command[0] in ['X', 'Y', 'Z']: # XR (rate) / YP (power) / ZS (speed)
            pen[{'X': 'rate', 'Y': 'power', 'Z': 'speed'}[command[0]]] = int(command[2:])
            if pen['down']:
                if len(paths[-1]['path']) < 2:
                    del paths[-1]
                paths.append({'params': {'rate':  pen['rate'],
                                         'power': pen['power'],
                                         'speed': pen['speed']},
                              'path': [[pen['x']/dpi, pen['y']/dpi]]})
    if len(paths[-1]['path']) < 2:
        del paths[-1]

    return paths

def paths2epi(paths, dpi = 600., headers = None, header = None, footer = None, jobname = 'unknown', autofocus = 1):
    if headers is not None:
        header = headers
        footer = headers
    epi = ''
    pen = {'x': None, 'y': None, 'rate': None, 'power': None, 'speed': None}
    params = {'rate': 'XR', 'power': 'YP', 'speed': 'ZS'}
    for path in paths:
        # Check if any parameters change
        for param in params:
            if path['params'][param] != pen[param]:
                epi += params[param]+str(path['params'][param])+';'
                pen[param] = path['params'][param]

        # Write the path
        if path['path'][0] != [pen['x'], pen['y']]:
            epi += 'PU%0.0f,%0.0f;'%(path['path'][0][0]*dpi, path['path'][0][1]*dpi)
        for i in range(1, len(path['path'])):
            epi += 'PD%0.0f,%0.0f;'%(path['path'][i][0]*dpi, path['path'][i][1]*dpi)
        pen['x'] = path['path'][-1][0]
        pen['y'] = path['path'][-1][1]

    if header:
        header = "%-12345X@PJL JOB NAME=" + jobname + "\r\nE@PJL ENTER LANGUAGE=PCL\r\n"+\
                 "&y" + str(autofocus) + "A&l0U&l0Z&u600D*p0X*p0Y*t600R*r0F&y"+\
                 "50P&z50S*r6600T*r5100S*r1A*rC%1BIN;"
    else:
        header = ''
    if footer:
        footer = "%0B%1BPUE%-12345X@PJL EOJ \r\n"
    else:
        footer = ''
    
    return header + epi + footer

def camm2paths(commands, dpi = 1000.):
    """
    Build a list of paths from Camm Vinyl Cutter commands.
    Each path is {'params': {'speed':, 'force':},
                  'path': [[x,y], [x,y], ...]}
    """

    # Create command list
    start = commands.find('!ST1;')
    if start == -1:
        start = None
    else:
        start += 5
    commands = commands[start:]
    commands = re.split('\s*;\s*', commands)

    # Build paths
    pen = {'x': 0, 'y': 0,
           'force': 100, 'speed': 100,
           'down': False}
    paths = []
    for command in commands:
        if command == '': continue
        if command[0] == 'P':
            if (command[1] == 'D') and not pen['down']:
                # First PD: save initial parameters
                paths.append({'params': {'force': pen['force'],
                                         'speed': pen['speed']},
                              'path': [[pen['x']/dpi, pen['y']/dpi]]})
                pen['down'] = True
            if command[2:] != '': # Disregard PU; commands with no coordinates
                temp = command[2:].split(',')
                pen['x'] = int(temp[0])
                pen['y'] = int(temp[1])
            if command[1] == 'D': # PD
                paths[-1]['path'].append([pen['x']/dpi, pen['y']/dpi])
            else: # PU
                if (len(paths) > 0) and (len(paths[-1]['path']) < 2):
                    del paths[-1]
                pen['down'] = False
        elif command[0] in ['!', 'V']: # !FS (force) / VS (speed)
            if command[0] == '!':
                pen['force'] = int(command[3:])
            else:
                pen['speed'] = int(command[2:])
            if pen['down']:
                if len(paths[-1]['path']) < 2:
                    del paths[-1]
                paths.append({'params': {'force': pen['force'],
                                         'speed': pen['speed']},
                              'path': [[pen['x']/dpi, pen['y']/dpi]]})
    if len(paths[-1]['path']) < 2:
        del paths[-1]

    return paths

def paths2camm(paths, dpi = 1000., headers = None):
    camm = ''
    pen = {'x': None, 'y': None, 'force': None, 'speed': None}
    params = {'force': '!FS', 'speed': 'VS'}
    for path in paths:
        # Check if any parameters change
        for param in params:
            if path['params'][param] != pen[param]:
                camm += params[param]+str(path['params'][param])+';'
                pen[param] = path['params'][param]

        # Write the path
        if path['path'][0] != [pen['x'], pen['y']]:
            camm += 'PU%0.0f,%0.0f;'%(path['path'][0][0]*dpi, path['path'][0][1]*dpi)
        for i in range(1, len(path['path'])):
            camm += 'PD%0.0f,%0.0f;'%(path['path'][i][0]*dpi, path['path'][i][1]*dpi)
        pen['x'] = path['path'][-1][0]
        pen['y'] = path['path'][-1][1]

    if headers:
        header = "PA;PA;!ST1;"
    else:
        header = ''
    
    return header + camm

def svg2paths(xml, format_callback = None):
    """
     Convert an SVG 1.1 file to the internal path format.
     
     xml contains the xml data taken from an SVG file.
     format_callback is a callback function that returns
    a list with the parameter settings for the path
    element. The function takes two arguments. The first
    is a list of group ids to which the element belongs.
    The second is a dictionary of {parameter: value} pairs
    for the xml element.

     TODO
      * <path> element commands: [sSqQtTaA]
      * <rect>, <circle>, <ellipse>
    """

    from helpers.convert import svg2paths as helper
    
    if format_callback is None:
        # The default: empty callback
        format_callback = lambda x, y: {}

    svg = helper.parse_xml(xml)
    assert svg[0].upper() == 'SVG', "Not an SVG file."

    # parse width, height from svg tag
    (width, trash) = helper.parse_float(svg[1]['width'])
    (height, trash) = helper.parse_float(svg[1]['height'])

    groups = []
    max_xy = [-1, -1]

    svg[1]['id'] = ''
    paths = helper.parse_svg_group(svg, {'groups': groups,
                                         'max_xy': max_xy,
                                         'format': format_callback})
    """
    # Rescale paths
    scale = [width/max_xy[0], height/max_xy[1]]
    for path in paths:
        for point in path['path']:
            point[0] *= scale[0]
            point[1] *= scale[1]
    """
    return paths

def paths2svg(paths, headers = None):
    minmax = [1e10, -1e10, 1e10, -1e10]
    svg = ''
    for path in paths:
        svg += '<path d="M %f,%f'%(path['path'][0][0], path['path'][0][1])
        for i in range(1, len(path['path'])):
            vertex = path['path'][i]
            svg += ' L %f,%f'%(vertex[0], vertex[1])
            minmax[0] = min(minmax[0], vertex[0])
            minmax[1] = max(minmax[1], vertex[0])
            minmax[2] = min(minmax[2], vertex[1])
            minmax[3] = max(minmax[3], vertex[1])
        svg += '"/>\n'
    if headers:
        header = '<?xml version="1.0" encoding="UTF-8"?>\n' +\
                 '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 1.1//EN" ' +\
                 '"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd">\n'
    else:
        header = ''
    header += '<svg width="%f" height="%f">\n<g style="stroke:rgb(0,0,0);fill:none">' %\
              (minmax[1]-minmax[0], minmax[3]-minmax[2])
    footer = '</g></svg>'
        
    return header + svg + footer

def paths2gcode(paths, pu = 0., pd = 0., headers = None, header = None, footer = None):
    if headers is not None:
        header = headers
        footer = headers

    dpi = 1.
    gcode = ''
    pen = {'x': None, 'y': None}
    for path in paths:
        # Write the path
        if path['path'][0] != [pen['x'], pen['y']]:
            gcode += 'G0 Z%f\n'%(pu,)
            gcode += 'X%f Y%f\n'%(path['path'][0][0]*dpi, path['path'][0][1]*dpi)
            gcode += 'G1 Z%f\n'%(path['params'].get('z', pd),)
        for i in range(1, len(path['path'])):
            gcode += 'X%f Y%f\n'%(path['path'][i][0]*dpi, path['path'][i][1]*dpi)
        pen['x'] = path['path'][-1][0]
        pen['y'] = path['path'][-1][1]

    if header:
        header = "%\n"
        header += "G20\n" # set units to inches
        header += "G90 G54 G28\n" # set absolute coordinates mode, use coord. sys. 1, go home
        header += "F10\n"    # spindle speed, spin clockwise, feedrate
    else:
        header = ''
    if footer:
        footer = ""
        footer += "G0 Z%f\n"%(pu,)  # pen up
        footer += "G90 G28\n"       # go home
        footer += "M30\n"           # stop program
        footer += "%\n"
    else:
        footer = ''
    
    return header + gcode + footer
