"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# REMOVE: chair for 0.5in stock, scale 1.9231
from __future__ import division
from convert import svg2paths, paths2svg
from render import render_paths
import sys

params = {
    'colors': {
        'selected': '#FF0000',
        'deselected': '#000000',
    },
    'scale': 40,
    'cursor': None, # the cursor used for block selection
}

# Read options
argc = 1
while (argc < len(sys.argv)) and (sys.argv[argc][0] == '-'):
    if sys.argv[argc][1] == 's': # scale
        params['scale'] *= eval(sys.argv[argc][2:])
    else:
        print "Warning: Unknown option '%s'"%(sys.argv[argc],)
    argc += 1
if argc == len(sys.argv):
    print "Error: No filename specified"
    sys.exit(1)

# Load SVG data
fp = file(sys.argv[argc])
svg = fp.read()
fp.close()
paths = svg2paths(svg)

# Offset to (0, 0)
minxy = [1e10, 1e10]
for path in paths:
    for vertex in path['path']:
        if vertex[0] < minxy[0]: minxy[0] = vertex[0]
        if vertex[1] < minxy[1]: minxy[1] = vertex[1]
for path in paths:
    for vertex in path['path']:
        vertex[0] -= minxy[0]
        vertex[1] -= minxy[1]
display = render_paths(paths, scale = params['scale'], title = "SVG Editor", grid = None,
                       format_callback = lambda x: params['colors']['deselected'])

# Map path canvas ids to paths
canvas_ids = {}
for path in paths:
    canvas_ids[path['params']['canvas_id']] = path

# Helper functions

# Remove deleted paths
def clear_deleted():
    global paths
    for i in range(len(paths)-1, -1, -1):
        if paths[i]['params'].get('deleted', False):
            del paths[i]

# Build quad tree
# Quadrants:
#     (+)
#    1 | 0
#   ---+--- (+)
#    3 | 2
def build_quad_tree(root):
    global paths
    
    max_entries = 10

    if len(root['entries']) <= max_entries:
        return

    # make new quadrants
    mid = [(root['minmax'][1]+root['minmax'][0])/2,
           (root['minmax'][3]+root['minmax'][2])/2]
    root['children'] = [
        {'minmax': [mid[0], root['minmax'][1], mid[1], root['minmax'][3]], 'entries': []},
        {'minmax': [root['minmax'][0], mid[0], mid[1], root['minmax'][3]], 'entries': []},
        {'minmax': [mid[0], root['minmax'][1], root['minmax'][2], mid[1]], 'entries': []},
        {'minmax': [root['minmax'][0], mid[0], root['minmax'][2], mid[1]], 'entries': []}]

    # sort entries into quadrants
    for entry in root['entries']:
        path = abs(entry)-1
        start = [-1, 0][entry > 0]
        xy = paths[path]['path'][start]
        
        child = 0
        if xy[0] < mid[0]: child += 1
        if xy[1] < mid[1]: child += 2

        root['children'][child]['entries'].append(entry)
    
    del root['entries']

    # recurse
    for child in root['children']:
        build_quad_tree(child)

def map_canvas_coords(x, y):
    global display
    return (display.canvas.canvasx(x), display.canvas.canvasy(y))

# Define canvas click events (select/deselect)
def click_select(event):
    global display

    (x, y) = map_canvas_coords(event.x, event.y)
    items = display.canvas.find_overlapping(x-2, y-2, x+2, y+2)
    for item in items:
        display.canvas.addtag_withtag('selected', item)
        display.canvas.itemconfig(item, fill = params['colors']['selected'])
display.canvas.bind('<Button-1>', click_select)

def click_deselect(event):
    global display
    
    (x, y) = map_canvas_coords(event.x, event.y)
    items = display.canvas.find_overlapping(x-2, y-2, x+2, y+2)
    for item in items:
        display.canvas.dtag(item, 'selected')
        display.canvas.itemconfig(item, fill = params['colors']['deselected'])
display.canvas.bind('<Control-Button-1>', click_deselect)

# Block select/deselect
def place_cursor(x, y):
    global params, display
    
    params['cursor'] = (x, y)
    display.canvas.create_line([x-7, y, x+7, y], tag="cursor", dash=3, fill="#0000FF")
    display.canvas.create_line([x, y-7, x, y+7], tag="cursor", dash=3, fill="#0000FF")

def remove_cursor():
    global params
    
    params['cursor'] = None
    display.canvas.delete("cursor")

def click_select_block(event):
    global display, params
    
    (x, y) = map_canvas_coords(event.x, event.y)
    if params['cursor'] is None:
        place_cursor(x, y)
    else:
        x = [params['cursor'][0], x]
        y = [params['cursor'][1], y]
        x.sort()
        y.sort()
        items = display.canvas.find_enclosed(x[0], y[0], x[1], y[1])
        for item in items:
            display.canvas.addtag_withtag('selected', item)
            display.canvas.itemconfig(item, fill = params['colors']['selected'])
        remove_cursor()
display.canvas.bind('<Button-3>', click_select_block)

def click_deselect_block(event):
    global display, params
    
    (x, y) = map_canvas_coords(event.x, event.y)
    if params['cursor'] is None:
        place_cursor(x, y)
    else:
        x = [params['cursor'][0], x]
        y = [params['cursor'][1], y]
        x.sort()
        y.sort()
        items = display.canvas.find_enclosed(x[0], y[0], x[1], y[1])
        for item in items:
            display.canvas.dtag(item, 'selected')
            display.canvas.itemconfig(item, fill = params['colors']['deselected'])
        remove_cursor()
display.canvas.bind('<Control-Button-3>', click_deselect_block)

# Select all (a)
def select_all(event):
    global display

    display.canvas.addtag_all('selected')
    items = display.canvas.find_all()
    for item in items:
        display.canvas.itemconfig(item, fill = params['colors']['selected'])
display.root.bind('<Control-a>', select_all)

# Deselect all (A)
def deselect_all(event):
    global display

    display.canvas.dtag('selected', 'selected')
    items = display.canvas.find_all()
    for item in items:
        display.canvas.itemconfig(item, fill = params['colors']['deselected'])
display.root.bind('<Control-A>', deselect_all)

# Invert selection (i)
def invert_selection(event):
    global display

    items = display.canvas.find_all()
    for item in items:
        if 'selected' in display.canvas.gettags(item):
            display.canvas.dtag(item, 'selected')
            display.canvas.itemconfig(item, fill = params['colors']['deselected'])
        else:
            display.canvas.addtag_withtag('selected', item)
            display.canvas.itemconfig(item, fill = params['colors']['selected'])
display.root.bind('<Control-i>', invert_selection)

# Delete selected paths (d)
def delete_paths(event):
    global display, canvas_ids
    
    items = display.canvas.find_withtag('selected')
    for item in items:
        canvas_ids[item]['params']['deleted'] = True
    display.canvas.delete('selected')
display.root.bind('<Control-d>', delete_paths)

# Break selected paths (b)
def split_paths(event):
    # TODO
    global display
    
# Join selected paths with overlapping start and end points (j)
def join_paths(event):
    global display, paths, params

    def plot_path(path, format):
        global display, params
        
        coords = []
        for point in path['path']:
            for i in range(2):
                coords.append(point[i]*params['scale'])
        # Add canvas identifier for later manipulation
        return display.canvas.create_line(coords, fill = format)

    clear_deleted()

    # map selected paths to path indices and find range
    minmax = [1e10, -1e10, 1e10, -1e10]
    items = display.canvas.find_withtag('selected')
    entries = [] # entries to add to root of quad tree
    for i in range(len(paths)):
        if paths[i]['params']['canvas_id'] in items:
            #print i+1, paths[i]['path'][0], paths[i]['path'][-1]
            entries += [i+1, -i-1]
            for j in range(-1, 1):
                vertex = paths[i]['path'][j]
                if vertex[0] < minmax[0]: minmax[0] = vertex[0]
                if vertex[0] > minmax[1]: minmax[1] = vertex[0]
                if vertex[1] < minmax[2]: minmax[2] = vertex[1]
                if vertex[1] > minmax[3]: minmax[3] = vertex[1]

    # build quad tree
    quad_root = {'minmax': minmax, 'entries': entries}
    build_quad_tree(quad_root)

    def print_quad(node):
        if node.has_key('children'):
            for child in node['children']:
                print_quad(child)
        else:
            print node['minmax']
            for entry in node['entries']:
                remap_entry = remap.get(entry, entry)
                path = abs(remap_entry)-1
                start = [-1, 0][remap_entry > 0]
                print entry, paths[path]['path'][start]
            print ''

    def remap_inverse(remap, y):
        for x in remap:
            if remap[x] == y:
                return x
        return y
    
    # join segments
    stack = [quad_root] # stack of tree nodes left to check
    remap = {} # remapping of entries when joined
    #print '---------------------------------------------'
    #print_quad(quad_root)
    while len(stack) > 0:
        node = stack.pop()
        
        if node.has_key('children'): # not a leaf node
            stack += node['children']
            continue

        i = 0
        while i < len(node['entries'])-1:
            remap_i = remap.get(node['entries'][i], node['entries'][i])
            start_i = [-1, 0][remap_i > 0]
            entry_i = abs(remap_i)-1
            for j in range(i+1, len(node['entries'])):
                remap_j = remap.get(node['entries'][j], node['entries'][j])
                start_j = [-1, 0][remap_j > 0]
                entry_j = abs(remap_j)-1
                if entry_i != entry_j: # not on same path
                    if paths[entry_i]['path'][start_i] == paths[entry_j]['path'][start_j]:
                        # coincident, merge paths and update pointers
                        inverse_i = remap_inverse(remap, -remap_i)
                        inverse_j = remap_inverse(remap, -remap_j)
                        #print 'i = %i, j = %i, (%i->%i, %i->%i)'%(i, j, node['entries'][i], remap_i, node['entries'][j], remap_j)
                        if start_i == 0:
                            if start_j == 0: # head to head
                                paths[entry_i]['path'].reverse()
                                remap[inverse_i] = remap_i
                                remap[inverse_j] = -remap_i
                            else: # head to tail
                                paths[entry_i]['path'].reverse()
                                paths[entry_j]['path'].reverse()
                                remap[inverse_i] = remap_i
                                remap[inverse_j] = -remap_i
                        else:
                            if start_j == -1: # tail to tail
                                paths[entry_j]['path'].reverse()
                                remap[inverse_j] = remap_i
                            else: # tail to head
                                remap[inverse_j] = remap_i

                        # merge
                        paths[entry_i]['path'] += paths[entry_j]['path'][1:]

                        # delete old plots
                        del canvas_ids[paths[entry_i]['params']['canvas_id']]
                        display.canvas.delete(paths[entry_i]['params']['canvas_id'])
                        del canvas_ids[paths[entry_j]['params']['canvas_id']]
                        display.canvas.delete(paths[entry_j]['params']['canvas_id'])
                        
                        # make new plot
                        new_id = plot_path(paths[entry_i], params['colors']['selected'])
                        paths[entry_i]['params']['canvas_id'] = new_id
                        canvas_ids[new_id] = paths[entry_i]
                        display.canvas.addtag_withtag('selected', new_id)

                        # mark old path as deleted
                        paths[entry_j]['params']['deleted'] = True

                        remap.pop(node['entries'][i], None)
                        remap.pop(node['entries'][j], None)
                        del node['entries'][j]
                        del node['entries'][i]
                        i -= 1

                        #print remap
                        #print '---------------------------------------------'
                        #print_quad(quad_root)
    
                        break
            i += 1

display.root.bind('<Control-j>', join_paths)

# Orient selected clockwise/anticlockwise
def is_clockwise(path):
    from math import atan2, pi

    extreme = 0
    for i in range(1, len(path)):
        if (path[i][0] < path[extreme][0]) or \
           ((path[i][0] == path[extreme][0]) and (path[i][1] < path[extreme][1])):
            extreme = i
    vec1 = [path[extreme-1][0]-path[extreme][0], path[extreme-1][1]-path[extreme][1]]
    vec2 = [path[extreme+1][0]-path[extreme][0], path[extreme+1][1]-path[extreme][1]]
    cross = vec1[0]*vec2[1]-vec1[1]*vec2[0]
    return cross > 0

def orient(clockwise):
    global paths, display
    
    items = display.canvas.find_withtag('selected')
    for item in items:
        if is_clockwise(canvas_ids[item]['path']) != clockwise:
            canvas_ids[item]['path'].reverse()
    
def orient_clockwise(event):
    orient(True)
display.root.bind('<Control-o>', orient_clockwise)

def orient_anticlockwise(event):
    orient(False)
display.root.bind('<Control-O>', orient_anticlockwise)

# Save paths to SVG (s)
def save_to_file(event):
    global paths

    clear_deleted()
    
    fp = file(sys.argv[argc] + '.new', 'w')
    fp.write(paths2svg(paths, headers = True))
    fp.close()
display.root.bind('<Control-s>', save_to_file)

print """
SVG Editor

  Mouse-left:       Select path
  Mouse-right:      Deselect path
  Ctrl-mouse-left:  Select block of paths
  Ctrl-mouse-right: Deselect block of paths
  Ctrl-a:           Select all
  Ctrl-A:           Deselect all
  Ctrl-b:           Break selected
  Ctrl-d:           Delete selected
  Ctrl-i:           Invert selection
  Ctrl-j:           Join selected
  Ctrl-o:           Orient selected clockwise
  Ctrl-O:           Orient selected anti-clockwise
  Ctrl-s:           Save to SVG
  q: Quit
"""

display.root.mainloop()
