"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division
import sys
import Tkinter

class Display(Tkinter.Toplevel):
    def __init__(self, window = (1000, 650), title = ''):
        # Top level
        self.root = Tkinter.Tk()
        self.root.title(title)
        self.root.protocol('WM_DELETE_WINDOW', self.destroy)
        self.root.bind('q', self.destroy)
        self.root.bind('Q', self.destroy)
        
        self.canvasframe = Tkinter.Frame(self.root)
        self.xscrollbar = Tkinter.Scrollbar(self.canvasframe, orient=Tkinter.HORIZONTAL)
        self.xscrollbar.grid(row=1, column=0, sticky=Tkinter.E+Tkinter.W)
        self.yscrollbar = Tkinter.Scrollbar(self.canvasframe)
        self.yscrollbar.grid(row=0, column=1, sticky=Tkinter.N+Tkinter.S)
        
        self.canvas = Tkinter.Canvas(self.canvasframe, width=window[0], height=window[1],
                                     background='white', xscrollcommand=self.xscrollbar.set,
                                     yscrollcommand=self.yscrollbar.set)
        self.canvas.grid(row=0, column=0, sticky=Tkinter.N+Tkinter.S+Tkinter.E+Tkinter.W)
        self.canvas.configure(scrollregion=(0,0,window[0],window[1]))
        self.xscrollbar.config(command=self.canvas.xview)
        self.yscrollbar.config(command=self.canvas.yview)
        
        self.canvasframe.pack()

    def destroy(self, trash = None):
        sys.exit()


def render_paths(paths, canvas = None, scale = 40, grid = None, format_callback = None, \
                 window = (1000, 650), title = ''):
    """
    paths: internal paths structure
    scale: all values in paths multiplied by this, seeing as internal units are inches,
           this is simply the dpi.
    window: size of the window on screen
    title: window title
    grid: a 2-tuple containing the distances between the x and y gridlines
    format_callback: defines colours for drawing lines.
                     takes 1 argument, the path parameters dict.
                     returns a colour code of the form '#0369CF'
    """
    if canvas is None:
        display = Display(window, title)
        canvas = display.canvas
    else:
        display = None

    if format_callback is None:
        format_callback = lambda x: '#000000'

    # Find bounding box
    max_xy = [-1, -1]
    for path in paths:
        for point in path['path']:
            for i in range(2):
                xy = point[i] * scale
                if xy > max_xy[i]: max_xy[i] = xy
    max_xy[0] = max(max_xy[0], window[0])
    max_xy[1] = max(max_xy[1], window[1])

    # Draw grid lines
    if grid != None:
        for x in range(0, max_xy[0], grid[0]*scale):
            canvas.create_line([x, 0, x, max_xy[1]], tag="grid", dash=4, fill="#F0E030")
        for y in range(0, max_xy[1], grid[1]*scale):
            canvas.create_line([0, y, max_xy[0], y], tag="grid", dash=4, fill="#F0E030")

    # Plot paths
    for path in paths:
        format = format_callback(path['params'])
        coords = []
        for point in path['path']:
            for i in range(2):
                coords.append(point[i]*scale)
        # Add canvas identifier for later manipulation
        path['params']['canvas_id'] = canvas.create_line(coords, fill = format)

    if display is not None:
        # Setup scroll bars
        xscrollmin = 0
        xscrollmax = max_xy[0]
        yscrollmin = 0
        yscrollmax = max_xy[1]
        canvas.configure(scrollregion = (xscrollmin, yscrollmin, xscrollmax, yscrollmax))
        if xscrollmax == window[0]:
            display.xscrollbar.grid_forget()
        else:
            display.xscrollbar.grid(row = 1, column = 0, sticky = Tkinter.E+Tkinter.W)
        if yscrollmax == window[1]:
            display.yscrollbar.grid_forget()
        else:
            display.yscrollbar.grid(row = 0, column = 1, sticky = Tkinter.N+Tkinter.S)

        display.root.update()
        return display

def render_epi(commands, dpi = 40, canvas=None, window = (960, 480), title = 'Epilog output',\
               grid = (1, 1), format_callback = None):

    from convert import epi2paths
    
    def default_callback(params):
        maxpower = 100
        
        if params['power'] > maxpower:
            return "#000000"
        else:
            return "#%02x%02x%02x"%(((maxpower-params['power'])*192//maxpower,)*3)
            
    if format_callback is None:
        format_callback = default_callback

    return render_paths(epi2paths(commands), scale = dpi, window = window, canvas=canvas,\
                        title = title, grid = grid, format_callback = format_callback)

def render_camm(commands, dpi = 40, window = (960, 480), title = 'Camm output',\
                grid = (1, 1), format_callback = None):

    from convert import camm2paths

    def default_callback(params):
        maxforce = 200

        if params['force'] > maxforce:
            colour = "#000000"
        else:
            colour = "#%02x%02x%02x"%(((maxforce-params['force'])*192//maxforce,)*3)
        return colour
            
    if format_callback is None:
        format_callback = default_callback

    return render_paths(camm2paths(commands), scale = dpi, window = window,\
                        title = title, grid = grid, format_callback = format_callback)

def render_rml(commands):
    import re

    def bound(bounding_box, point):
        if point[0] < bounding_box[0]:
            bounding_box[0] = point[0]
        if point[0] > bounding_box[1]:
            bounding_box[1] = point[0]
        if point[1] < bounding_box[2]:
            bounding_box[2] = point[1]
        if point[1] > bounding_box[3]:
            bounding_box[3] = point[1]
        if len(point) > 2:
            if point[2] < bounding_box[4]:
                bounding_box[4] = point[2]
            if point[2] > bounding_box[5]:
                bounding_box[5] = point[2]

    def find_closest(sorted_list, val, a = 0, b = None):
        if b == None:
            b = len(sorted_list)

        if b == a+2:
            return [a+1, a][val - sorted_list[a] < sorted_list[a+1] - val]
        elif b == a+1:
            return a

        mid = a + (b-a)//2
        if val > sorted_list[mid]:
            return find_closest(sorted_list, val, mid, b)
        else:
            return find_closest(sorted_list, val, a, mid+1)
            
    def update_grid(grid_x, grid_y, grid_z, pos0, pos1):
        if pos0[:2] == pos1[:2]:
            x = find_closest(grid_x, pos0[0])
            y = find_closest(grid_y, pos0[1])
            grid_z[x, y] = min(grid_z[x, y], pos0[2], pos1[2])
            return
        
        grid_xy = [grid_x, grid_y]
        
        # Find along which axis the change is greatest
        major = int(abs(pos0[0]-pos1[0]) < abs(pos0[1]-pos1[1]))
        minor = 1-major
        dminor = (pos0[minor]-pos1[minor])/(pos0[major]-pos1[major])
        
        if pos0[major] > pos1[major]:
            temp = pos0
            pos0 = pos1
            pos1 = temp

        index = [-1, -1]
        index[major] = find_closest(grid_xy[major], pos0[major])
        dz = (pos1[2]-pos0[2])/(pos1[major]-pos0[major])
        height = pos0[2]

        while (index[major] < len(grid_xy[major])) and \
              (grid_xy[major][index[major]] <= pos1[major]):
            index[minor] = find_closest(grid_xy[minor],
                                        pos0[minor]+dminor*(grid_xy[major][index[major]]-
                                                            pos0[major]))
            grid_z[index[0], index[1]] = min(grid_z[index[0], index[1]], height)
            index[major] += 1
            height += dz


    DPI = 1000. # TODO: read this from the file header

    position = [0., 0.]
    bounding_box = [1e10, -1e10, 1e10, -1e10, 1e10, -1e10]

    start = commands.find('!MC1;') + 5
    stop  = commands.rfind('!MC0')

    commands = commands[start:stop-1].strip()
    commands = re.split(';\s*', commands)

    # Find bounding box
    for command in commands:
        if command[0] == 'P':
            if command[1] == 'D':
                bound(bounding_box, position)
            temp = command[2:].split(',')
            position[0] = float(temp[0])/DPI
            position[1] = float(temp[1])/DPI
            if command[1] == 'D':
                bound(bounding_box, position)
        elif command[0:3] == '!PZ':
            temp = command[3:].split(',')
            z_down = float(temp[0])/DPI
            z_up   = float(temp[1])/DPI
            if z_down < bounding_box[4]:
                bounding_box[4] = z_down
            if z_up > bounding_box[5]:
                bounding_box[5] = z_up
        elif command[0] == 'Z':
            temp = command[1:].split(',')
            assert len(temp)%3 == 0, "Number of coordinates (x,y,z) not a multiple of 3."
            for i in range(0, len(temp), 3):
                position = [float(temp[i])/DPI, float(temp[i+1])/DPI, float(temp[i+2])/DPI]
                bound(bounding_box, position)


    import scipy
    import Gnuplot
    gp = Gnuplot.Gnuplot(persist = 1)
    gp('unset autoscale')
    #gp('set size ratio -1')
    gp('set dat style lines')
    gp('set xrange [' + str(bounding_box[0]) + ':' + str(bounding_box[1]) + ']')
    gp('set yrange [' + str(bounding_box[2]) + ':' + str(bounding_box[3]) + ']')
    gp('set zrange [' + str(bounding_box[4]) + ':' + str(bounding_box[5]) + ']')

    grid_points = 30
    grid_x = scipy.linspace(bounding_box[0], bounding_box[1], grid_points)
    grid_y = scipy.linspace(bounding_box[2], bounding_box[3], grid_points)
    grid_z = scipy.ones([grid_points, grid_points], typecode='d') * 1.

    position = [0., 0., 0.]
    pen_height = {'up': 0., 'down': 0.}
    for command in commands:
        if command[0] == 'P':
            height = pen_height[{'U': 'up', 'D': 'down'}[command[1]]]
            old_position = position[:2]
            temp = command[2:].split(',')
            position = [float(temp[0])/DPI, float(temp[1])/DPI, height]
            update_grid(grid_x, grid_y, grid_z, old_position + [height], position)
        elif command[0:3] == '!PZ':
            temp = command[3:].split(',')
            pen_height['down'] = float(temp[0])/DPI
            pen_height['up']   = float(temp[1])/DPI
        elif command[0] == 'Z':
            temp = command[1:].split(',')
            assert len(temp)%3 == 0, "Number of coordinates (x,y,z) not a multiple of 3."
            for i in range(0, len(temp), 3):
                old_position = position[:]
                position = [float(temp[i])/DPI, float(temp[i+1])/DPI, float(temp[i+2])/DPI]
                update_grid(grid_x, grid_y, grid_z, old_position, position)

    gp.splot(Gnuplot.PlotItems.GridData(grid_z, grid_x, grid_y))

    return None
