"""
Copyright (c) 2008, Carl Scheffler
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * The name of the copyright owner may not be used to endorse or promote
      products derived from this software without specific prior written
      permission.

THIS SOFTWARE IS PROVIDED BY <copyright holder> ``AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import division

_params = {
    'bezier_segments': 20, # number of segments with which to approximate bezier curves
}

def skip_chars(chars, text, start = 0):
    """
     Skip characters in the string chars and return the
    index of the first character not in chars. Return
    len(text) if text[start:] are all in chars.
    """
    while (start < len(text)) and (text[start] in chars):
        start += 1
    return start

def skip_whitespace(text, start = 0):
    """
     Skip whitespace in text. Return position of first
    non-whitespace character or len(text) if text[start:]
    is all whitespace.
    """
    return skip_chars(' \r\n\t', text, start)

def parse_float(text, start = 0):
    """
     Parse a float from text.
     Return (<float>, stop) where stop is the index of the
    first character after the float. If no float was found
    stop == start.
    """
    stop = start
    if (stop < len(text)) and (text[stop] == '-'):
        stop += 1
    while (stop < len(text)) and (text[stop] in '0123456789.eE'):
        stop += 1
    if start == stop: # no float found
        return (0., start)
    else:
        return (float(text[start:stop]), stop)

def parse_coords(text, start = 0):
    """
     Parse a coordinate list from text.
     Return (<list of floats>, stop) where stop is the index
    of the first non-whitespace character after the coordinate
    list.
    """
    coords = []
    start = skip_whitespace(text, start)
    while True:
        (f, stop) = parse_float(text, start)
        if start == stop:
            return (coords, stop)
        coords.append(f)
        start = skip_chars(', \r\n\t', text, stop)

def parse_parameters(text, start = 0, stop = None):
    import re
    return dict(re.findall(r'([a-zA-Z][0-9a-zA-Z:]*)\s*=\s*"([^"]*)"', text[start:stop]))

def end_of_tag(text, start = 0):
    # Find (position of the closing >)+1 for a tag that was opened before start
    opening = text.find('<', start)
    closing = text.find('>', start)

    while (opening < closing) and (opening != -1):
        closing = end_of_tag(text, opening+1)
        opening = text.find('<', closing)
        closing = text.find('>', closing)
    
    return closing+1

def parse_xml(xml):
    root = ['__root__', {}, []]
    open_tag = root
    parent = []

    pos = 0
    while pos < len(xml):
        # Whitespace
        if xml[pos] in ' \r\n\t':
            pos += 1
            continue

        if xml[pos] != '<':
            print "Warning: skipping text (%s)"%(xml[pos:pos+10],)
            pos = xml.find('<', pos)
        
        # Ignored tags
        if xml[pos+1] in '?!':
            pos = end_of_tag(xml, pos+1)
            assert pos != 0
            continue

        # Closing tags
        if xml[pos+1] == '/':
            stop = xml.find('>', pos)
            tag_name = xml[pos+2:stop]
            assert open_tag[0] == tag_name, \
                   "Invalid closing tag '%s' in '%s'"%(tag_name, open_tag[0])
            open_tag = parent.pop()
            pos = stop+1
            continue

        # Opening and one-line tags
        stop = pos+1
        while True:
            char = xml[stop].upper()
            if ((char < 'A') or (char > 'Z')) and (char not in ':'):
                break
            stop += 1
        tag_name = xml[pos+1:stop]
        pos = stop
        stop = end_of_tag(xml, pos)
        tag_data = xml[pos:stop-1]
        pos = stop
        tag_opening = (len(tag_data) == 0) or (tag_data[-1] != '/')

        # Create new entry
        open_tag[2].append([tag_name, parse_parameters(tag_data)])
        if tag_opening:
            parent.append(open_tag)
            open_tag[2][-1].append([])
            open_tag = open_tag[2][-1]

    assert open_tag is root, "Not all tags closed at EOF"
    return root[2][0]

def add_path(path, svg_params, element_params):
    if len(path) < 4:
        return []

    result = {'params': svg_params['format'](svg_params['groups'], element_params), 'path': []}
    for i in range(0, len(path), 2):
        # Update bounding box
        if path[i] > svg_params['max_xy'][0]: svg_params['max_xy'][0] = path[i]
        if path[i+1] > svg_params['max_xy'][1]: svg_params['max_xy'][1] = path[i+1]

        # Add point
        result['path'].append([path[i], path[i+1]])
    return [result]

def bezier_path(control):
    global _params

    path = control[0][:]
    for seg in range(1, _params['bezier_segments']):
        t = seg/_params['bezier_segments'] # interpolation parameter

        control1 = control
        while len(control1) > 1:
            control2 = []
            for i in range(1, len(control1)):
                control2.append([control1[i][0]*t + control1[i-1][0]*(1-t),
                                 control1[i][1]*t + control1[i-1][1]*(1-t)])
            control1 = control2

        path += control1[0]
    path += control[3]
    return path

def parse_svg_path(data, svg_params, element_params):
    result = []
    
    pen = [0., 0.]
    path = pen[:]
    pos = 0
    while pos < len(data):
        # Skip whitespace
        pos = skip_whitespace(data, pos)
        if pos == len(data): break

        command = data[pos]
        assert (command.upper() >= 'A') and (command.upper() <= 'Z'),\
               "Invalid command '%s'"%(command,)

        (coords, pos) = parse_coords(data, pos+1)
        
        if command in 'mM': # Move to
            result += add_path(path, svg_params, element_params)
            if command == 'M': # absolute
                path = coords[:]
                pen = path[-2:]
            else: # relative
                path = []
                for i in range(0, len(coords), 2):
                    pen[0] += coords[i]
                    pen[1] += coords[i+1]
                    path += pen
        elif command in 'zZ': # Close path
            # check if next command is a 'move to'
            if (pos < len(data)) and (data[pos] in 'Mm'):
                pen = path[-2:]
            else:
                pen = path[:2]
            if len(path) > 4:
                path += path[:2]
            result += add_path(path, svg_params, element_params)
            path = pen[:]
        elif command in 'lL': # Line to
            if command == 'L': # absolute
                path += coords
                pen = path[-2:]
            else: # relative
                for i in range(0,len(coords), 2):
                    pen[0] += coords[i]
                    pen[1] += coords[i+1]
                    path += pen
        elif command in 'hH': # Horizontal line to
            if command == 'H': # absolute
                for i in range(len(coords)):
                    path += [coords[i], pen[1]]
                pen[0] = coords[-1]
            else: # relative
                for i in range(len(coords)):
                    pen[0] += coords[i]
                    path += pen
        elif command in 'vV': # Vertical line to
            if command == 'V': # absolute
                for i in range(len(coords)):
                    path += [pen[0], coords[i]]
                pen[1] = coords[-1]
            else: # relative
                for i in range(len(coords)):
                    pen[1] += coords[i]
                    path += pen
        elif command in 'cC': # Cubic bezier curve
            if command == 'C': # absolute
                for i in range(0, len(coords), 6):
                    control = [pen[:]]
                    for j in range(0, 6, 2):
                        control.append([coords[i+j], coords[i+j+1]])
                    path += bezier_path(control)
                    pen = control[-1]
            else: # relative
                for i in range(0, len(coords), 6):
                    control = [pen[:]]
                    for j in range(0, 6, 2):
                        control.append([pen[0]+coords[i+j], pen[1]+coords[i+j+1]])
                    path += bezier_path(control)
                    pen = control[-1]
        else:
            print coords
            assert False, "Unknown path command '%s'"%(command,)

    result += add_path(path, svg_params, element_params)
    return result

def parse_svg_group(group, svg_params):
    paths = []
    svg_params['groups'].append(group[1].get('id', '')) # Push group id onto stack
    
    for element in group[2]:
        if element[0] == 'g': # sub-group
            paths += parse_svg_group(element, svg_params)
            
        elif element[0] == 'path':
            paths += parse_svg_path(element[1]['d'], svg_params, element[1])
            
        elif element[0] == 'polyline':
            (coords, pos) = parse_coords(element[1]['points'])
            assert pos == len(element[1]['points']),\
                   "Length mismatch (%i, %s)"%(pos, element[1]['points'])
            paths += add_path(coords, svg_params, element[1])
            
        elif element[0] == 'line':
            paths += add_path([float(element[1]['x1']),
                               float(element[1]['y1']),
                               float(element[1]['x2']),
                               float(element[1]['y2'])], svg_params, element[1])
        elif element[0] == 'rect':
            x = float(element[1]['x'])
            y = float(element[1]['y'])
            w = float(element[1]['width'])
            h = float(element[1]['height'])
            paths += add_path([x, y, x+w, y, x+w, y+h, x, y+h, x, y],
                              svg_params, element[1])
            
        elif element[0] == 'polygon':
            (coords, pos) = parse_coords(element[1]['points'])
            assert pos == len(element[1]['points']),\
                   "Length mismatch (%i, %s)"%(pos, element[1]['points'])
            coords += coords[:2]
            paths += add_path(coords, svg_params, element[1])
            
        else:
            assert element[0] in ['metadata', 'defs'],\
                   "Unknown element '%s'"%(element[0],)

    del svg_params['groups'][-1] # Pop group id off stack
    return paths
